# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class GCPNodeDisk(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'boot': 'bool',
        'auto_delete': 'bool',
        'type': 'str',
        'initialize_params': 'object'
    }

    attribute_map = {
        'boot': 'boot',
        'auto_delete': 'autoDelete',
        'type': 'type',
        'initialize_params': 'initializeParams'
    }

    def __init__(self, boot=None, auto_delete=None, type=None, initialize_params=None, local_vars_configuration=None):  # noqa: E501
        """GCPNodeDisk - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._boot = None
        self._auto_delete = None
        self._type = None
        self._initialize_params = None
        self.discriminator = None

        if boot is not None:
            self.boot = boot
        if auto_delete is not None:
            self.auto_delete = auto_delete
        if type is not None:
            self.type = type
        if initialize_params is not None:
            self.initialize_params = initialize_params

    @property
    def boot(self):
        """Gets the boot of this GCPNodeDisk.  # noqa: E501


        :return: The boot of this GCPNodeDisk.  # noqa: E501
        :rtype: bool
        """
        return self._boot

    @boot.setter
    def boot(self, boot):
        """Sets the boot of this GCPNodeDisk.


        :param boot: The boot of this GCPNodeDisk.  # noqa: E501
        :type: bool
        """

        self._boot = boot

    @property
    def auto_delete(self):
        """Gets the auto_delete of this GCPNodeDisk.  # noqa: E501


        :return: The auto_delete of this GCPNodeDisk.  # noqa: E501
        :rtype: bool
        """
        return self._auto_delete

    @auto_delete.setter
    def auto_delete(self, auto_delete):
        """Sets the auto_delete of this GCPNodeDisk.


        :param auto_delete: The auto_delete of this GCPNodeDisk.  # noqa: E501
        :type: bool
        """

        self._auto_delete = auto_delete

    @property
    def type(self):
        """Gets the type of this GCPNodeDisk.  # noqa: E501


        :return: The type of this GCPNodeDisk.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this GCPNodeDisk.


        :param type: The type of this GCPNodeDisk.  # noqa: E501
        :type: str
        """

        self._type = type

    @property
    def initialize_params(self):
        """Gets the initialize_params of this GCPNodeDisk.  # noqa: E501


        :return: The initialize_params of this GCPNodeDisk.  # noqa: E501
        :rtype: object
        """
        return self._initialize_params

    @initialize_params.setter
    def initialize_params(self, initialize_params):
        """Sets the initialize_params of this GCPNodeDisk.


        :param initialize_params: The initialize_params of this GCPNodeDisk.  # noqa: E501
        :type: object
        """

        self._initialize_params = initialize_params

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GCPNodeDisk):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GCPNodeDisk):
            return True

        return self.to_dict() != other.to_dict()
